import inside from 'is-path-inside';
import { join, extname, toUnix, basename, relative, dirname, removeExt } from 'upath';
import chalk from 'chalk';
import superjson from 'superjson';
export { default as superjson } from 'superjson';
import consola from 'consola';
export { default as consola } from 'consola';
import path, { join as join$1, resolve } from 'path';
import { sync } from 'pkg-dir';
import createJITI from 'jiti';
import _, { noop, kebabCase } from 'lodash';
import { createConfiguration } from '@midwayjs/core';
import { __decorate } from 'tslib';
import { Inject, Provide, Controller, Get, Post } from '@midwayjs/decorator';
import { als as als$1 } from 'asynchronous-local-storage';
import { existsSync, readFileSync } from 'fs';
import { sync as sync$1 } from 'globby';
import staticCache from 'koa-static-cache';
export { defaults, request } from '../request/dist/midwayjs-hooks-core-request.esm.js';
import { init, parse } from 'es-module-lexer';
import art from 'art-template';
import prettier from 'prettier';
import 'axios';

class ServerRouter {
  constructor(root, config, useSource = true) {
    this.root = void 0;
    this.config = void 0;
    this.routes = new Map();
    this.useSource = void 0;
    this.root = root;
    this.config = config;
    this.useSource = useSource;
  } // src/apis


  get source() {
    if (this.useSource) {
      return join(this.root, this.config.source);
    }

    return join(this.root, this.config.build.outDir);
  }

  getRouteConfig(file) {
    const {
      routes
    } = this.config;
    const index = routes.findIndex(route => {
      const dir = this.getApiDirectory(route.baseDir);
      return this.inside(file, dir);
    });

    if (!routes[index]) {
      return null;
    }

    return {
      underscore: false,
      ...routes[index]
    };
  } // src/apis/lambda


  getApiDirectory(baseDir) {
    return join(this.source, baseDir);
  }

  isApiFile(file) {
    if (file.endsWith('.test.ts') || file.endsWith('.test.js')) {
      return false;
    }

    if (extname(file) !== '.ts' && extname(file) !== '.js') {
      return false;
    }

    const route = this.getRouteConfig(file);
    return !!route;
  }

  inside(child, parent) {
    return inside(toUnix(child), toUnix(parent));
  }

  getBaseUrl(file) {
    const url = this.getHTTPPath(file, '', true);

    if (url === '/*') {
      return '/';
    }

    if (url.endsWith('/*')) {
      return url.slice(0, -2);
    }

    return url;
  }

  getHTTPPath(file, method, isExportDefault) {
    const {
      basePath,
      baseDir,
      underscore
    } = this.getRouteConfig(file);
    const lambdaDirectory = this.getApiDirectory(baseDir);
    const {
      isCatchAllRoutes,
      filename
    } = parseFilename(basename(file, extname(file)));
    const fileRoute = filename === 'index' ? '' : filename; // TODO remove underscore support in future

    const ApiMethodPrefix = '_';
    const methodPrefix = underscore ? ApiMethodPrefix : '';
    const func = isExportDefault ? '' : `${methodPrefix}${method}`;
    const api = toUnix(join(basePath,
    /**
     * /apis/lambda/index.ts -> ''
     * /apis/lambda/todo/index.ts -> 'todo'
     */
    relative(lambdaDirectory, dirname(file)),
    /**
     * index -> ''
     * demo -> '/demo'
     */
    fileRoute, // getTodoList -> _getTodoList
    func, isCatchAllRoutes ? '/*' : ''));
    /**
     * duplicate routes
     */

    const existPath = this.routes.get(api);

    if (existPath && existPath !== toUnix(file)) {
      duplicateLogger(this.root, existPath, file, api);
    }

    this.routes.set(api, toUnix(file));
    return api;
  }

}

function parseFilename(filename) {
  var _re$exec;

  const re = /\[\.{3}(.+)]/;
  const isCatchAllRoutes = re.test(filename);
  return {
    isCatchAllRoutes,
    filename: isCatchAllRoutes ? (_re$exec = re.exec(filename)) === null || _re$exec === void 0 ? void 0 : _re$exec[1] : filename
  };
}

function duplicateLogger(root, existPath, currentPath, api) {
  consola.info('[ %s ] Duplicate routes detected. %s and %s both resolve to %s. Reference: %s', chalk.yellow('warn'), chalk.cyan(relative(root, existPath)), chalk.cyan(relative(root, currentPath)), chalk.cyan(api), 'https://www.yuque.com/midwayjs/faas/et7x4k');
}

function getProjectRoot(cwd) {
  return sync(cwd) || process.cwd();
}
function getConfig(cwd) {
  const root = getProjectRoot(cwd);
  const configs = {
    ts: path.join(root, 'midway.config.ts'),
    js: path.join(root, 'midway.config.js')
  };
  const userConfig = tryRequire(configs.ts) || tryRequire(configs.js);
  return _.defaultsDeep({}, userConfig, {
    superjson: false,
    source: './src/apis',
    build: {
      viteOutDir: './build',
      outDir: './dist'
    },
    request: {
      client: '@midwayjs/hooks-core/request'
    }
  });
}
function defineConfig(config) {
  return config;
}

const tryRequire = id => {
  try {
    const jiti = createJITI();
    const contents = jiti(id);
    if ('default' in contents) return contents.default;
    return contents;
  } catch {
    return undefined;
  }
};

function isDevelopment() {
  if (process.env.MIDWAY_TS_MODE === 'true' || process.env.NODE_ENV === 'test' || process.env.NODE_ENV === 'development') {
    return true;
  }

  return false;
}

/**
 * @private
 * private api, may change without notice.
 * Use asynchronous-local-storage due to serverless environment does not support node.js 12.17.0
 */
const als = {
  getStore(key) {
    return als$1.get(key);
  },

  run(ctx, callback) {
    return new Promise(resolve => {
      als$1.runWith(async () => {
        resolve(await callback());
      }, ctx);
    });
  }

};

class HooksDevComponent {
  constructor(config) {
    this.internal = void 0;
    this.runtime = void 0;
    this.router = void 0;
    this.root = void 0;
    this.container = void 0;

    this.useAsyncLocalStorage = async (ctx, next) => {
      const enableSuperjson = this.internal.superjson;
      await als.run({
        ctx
      }, async () => {
        try {
          await next();
        } catch (error) {
          if (enableSuperjson) {
            ctx.status = 500;
            ctx.body = superjson.serialize(error);
          } else {
            throw error;
          }
        }
      });
    };

    this.runtime = config.runtime;
    this.internal = config.internal;
    this.router = config.router;
    this.root = config.root;
  }

  init() {
    const {
      routes
    } = this.internal;
    let count = 0;
    const totalCount = routes.reduce((totalCount, route) => {
      const dir = join$1(this.router.source, route.baseDir);
      const files = sync$1([dir]).filter(file => this.router.isApiFile(file));
      return totalCount + files.length;
    }, 0);
    const configuration = createConfiguration({
      namespace: '@midwayjs/hooks',
      directoryResolveFilter: routes.map(route => {
        return {
          pattern: join$1(this.router.source, route.baseDir),
          ignoreRequire: true,
          filter: (_, file, container) => {
            if (!this.router.isApiFile(file)) {
              return;
            }

            this.container = container;

            const mod = require(file);

            this.createApi(mod, file);
            this.container.bindClass(mod, '', file);
            count++;

            if (count === totalCount) {
              this.afterCreate();
            }
          }
        };
      })
    });
    configuration.onReady((_, app) => {
      this.useGlobalMiddleware(app);
    }).onStop(noop);
    return {
      Configuration: configuration
    };
  }

  afterCreate() {}

  createApi(mod, file) {
    var _mod$config;

    const modMiddleware = (mod === null || mod === void 0 ? void 0 : (_mod$config = mod.config) === null || _mod$config === void 0 ? void 0 : _mod$config.middleware) || [];
    Object.keys(mod).filter(key => typeof mod[key] === 'function').forEach(key => {
      this.createApiFunction({
        fn: mod[key],
        file: file,
        isExportDefault: key === 'default',
        modMiddleware
      });
    });
  }

  createApiFunction(config) {
    const {
      fn,
      file,
      isExportDefault,
      modMiddleware
    } = config;
    const fnName = isExportDefault ? '$default' : fn.name;
    const id = this.getFunctionId(file, fnName, isExportDefault);
    const containerId = 'hooks::' + id;
    const httpPath = this.router.getHTTPPath(file, fnName, isExportDefault);
    const httpMethod = fn.length === 0 ? 'GET' : 'POST'; // Set param for unit testing

    fn._param = {
      url: httpPath,
      method: httpMethod,
      meta: {
        functionName: id
      }
    }; // Apply mod middleware

    fn.middleware = (fn.middleware || (fn.middleware = [])).concat(modMiddleware).map(this.useHooksMiddleware);
    this.registerApiFunction({
      containerId,
      httpMethod,
      httpPath,
      fn
    });
  }

  getFunctionId(file, functionName, isExportDefault) {
    const relativePath = relative(this.router.source, file); // src/apis/lambda/index.ts -> apis-lambda-index

    const id = kebabCase(removeExt(relativePath, extname(relativePath)));
    const name = [id, isExportDefault ? '' : `-${functionName}`].join('');
    return name.toLowerCase();
  }

  registerApiFunction(config) {
    const {
      containerId,
      httpMethod,
      httpPath,
      fn
    } = config;
    const Method = httpMethod === 'GET' ? Get : Post;
    const enableSuperjson = this.internal.superjson;
    let FunctionContainer = class FunctionContainer {
      constructor() {
        this.ctx = void 0;
      }

      async handler() {
        var _this$ctx$request, _this$ctx$request$bod;

        let args = ((_this$ctx$request = this.ctx.request) === null || _this$ctx$request === void 0 ? void 0 : (_this$ctx$request$bod = _this$ctx$request.body) === null || _this$ctx$request$bod === void 0 ? void 0 : _this$ctx$request$bod.args) || [];

        if (typeof args === 'string') {
          args = JSON.parse(args);
        }

        const response = await fn(...args);

        if (enableSuperjson) {
          return superjson.serialize(response);
        }

        return response;
      }

    };

    __decorate([Inject()], FunctionContainer.prototype, 'ctx', void 0);

    __decorate([Method(httpPath, {
      middleware: fn.middleware
    })], FunctionContainer.prototype, 'handler', null);

    FunctionContainer = __decorate([Provide(containerId), Controller('/')], FunctionContainer);
    this.container.bind(containerId, FunctionContainer);
  }

  useGlobalMiddleware(app) {
    app.use(this.useAsyncLocalStorage); // Apply global middleware from config

    if (Array.isArray(this.runtime.middleware)) {
      this.runtime.middleware.forEach(middleware => app.use(this.useHooksMiddleware(middleware)));
    }
  }

  useHooksMiddleware(fn) {
    return (...args) => {
      /**
       * Hooks middleware
       * const middleware = (next) => { const ctx = useContext() }
       */
      if (fn.length === 1) {
        const next = args[args.length - 1];
        return fn(next);
      }

      return fn(...args);
    };
  }

  isFullStackProject() {
    return [// Vite
    'vite.config.ts', 'vite.config.js', // build-scritps
    'build.json', 'build.js'].some(config => existsSync(join$1(this.root, config)));
  }

}

class HooksProductionComponent extends HooksDevComponent {
  useGlobalMiddleware(app) {
    super.useGlobalMiddleware(app);

    if (!this.isFullStackProject()) {
      return;
    }

    if (this.existRootPath) {
      return;
    }

    const baseDir = app.getBaseDir();
    const dist = join(baseDir, '..', this.internal.build.viteOutDir);
    const mw = staticCache({
      dir: dist,
      dynamic: true,
      alias: {
        '/': 'index.html'
      },
      buffer: true,
      gzip: true
    });
    app.use(mw);
  }

  afterCreate() {
    if (!this.isFullStackProject()) {
      return;
    }

    if (this.existRootPath) {
      consola.debug('The route `/` or `/*` is registered, you need to host the front-end page manually');
      return;
    }

    this.registerApiFunction({
      containerId: 'hooks:host',
      httpMethod: 'GET',
      httpPath: '/*',
      fn: async () => {}
    });
  }

  get existRootPath() {
    const {
      routes
    } = this.router;
    return routes.has('/') || routes.has('/*');
  }

}

/**
 * Create hooks component
 */

const hooks = (runtime = {}) => {
  const root = getProjectRoot();
  const internal = getConfig();
  const router = new ServerRouter(root, internal, isDevelopment());
  const Component = isDevelopment() ? HooksDevComponent : HooksProductionComponent;
  const cmp = new Component({
    runtime,
    internal,
    router,
    root
  });
  return cmp.init();
};

// TODO Refactor SDK generator to support custom request sdk
async function generate(baseUrl, code, superjson, sdk) {
  await init;
  const [, exports] = parse(code);

  if (exports.length === 0) {
    return null;
  }

  const funcs = exports.map(name => {
    return {
      url: getUrl(baseUrl, name),
      meta: {
        superjson
      },
      isExportDefault: name === 'default',
      name: name === 'default' ? '$default' : name
    };
  });
  const templates = {
    dev: resolve(__dirname, '../../template/request.art'),
    prod: resolve(__dirname, '../template/request.art')
  };
  const templatePath = existsSync(templates.dev) ? templates.dev : templates.prod;
  const template = readFileSync(templatePath, {
    encoding: 'utf-8'
  });
  return prettier.format(art.render(template, {
    funcs,
    SDK: sdk
  }), {
    semi: true,
    singleQuote: true,
    parser: 'babel'
  });
}

function getUrl(baseUrl, name) {
  if (name === 'default') {
    return baseUrl;
  }

  if (baseUrl.endsWith('/')) {
    return `${baseUrl}${name}`;
  }

  return `${baseUrl}/${name}`;
}

art.defaults.htmlMinifierOptions = {
  collapseWhitespace: false,
  minifyCSS: false,
  minifyJS: false,
  ignoreCustomFragments: []
};

art.defaults.imports.stringify = json => JSON.stringify(json, null, 2);

art.defaults.escape = false;

export { ServerRouter, als, defineConfig, generate, getConfig, getProjectRoot, hooks, isDevelopment };
